import os
import webbrowser
from pathlib import Path
from argparse import ArgumentParser

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles

from .routers import solutions
from .config import config

# Initialize default app
app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# Initialize api app
api_app = FastAPI()
api_app.include_router(solutions.router)
app.mount("/api", api_app)

# Mount explorer as static files
explorer_path = os.path.join(os.path.dirname(__file__), "explorer")
explorer_url = "/"
app.mount(
    explorer_url, StaticFiles(directory=explorer_path, html=True), name="explorer"
)


def start_server(
    solution_folder: str,
    port: int,
    app_name: str | None = None,
    api_url: str | None = None,
    significant_digits: int | None = None,
    reload: bool = False,
    no_open_browser: bool = False,
    fd: int | None = None,
) -> None:
    if api_url is None:
        api_url = f"http://127.0.0.1:{port}/api/"
    if app_name is None:
        app_name = ""

    config.SOLUTION_FOLDER = solution_folder
    if significant_digits is not None:
        config.RESPONSE_SIGNIFICANT_DIGITS = significant_digits

    env_file = Path(__file__).parent / "explorer" / "_app" / "env.js"
    with open(env_file, "w") as file:
        file.write(
            f'export const env={{"PUBLIC_TEMPLE_URL":"{api_url}", "PUBLIC_APP_NAME":"{app_name}"}}'
        )

    # Start the uvicorn server
    if not no_open_browser:
        webbrowser.open(f"http://localhost:{port}/", new=2)
    uvicorn.run("zen_temple.main:app", host="localhost", port=port, log_level="info", reload=reload, fd=fd)


def find_outputs_folder(outputs_folder: str | None) -> str:
    """
    Verify if the outputs folder exists. Otherwise, goes through a list of default paths.
    If none of the default paths exist, it raises an error.
    """
    if outputs_folder is not None:
        outputs_path = Path(outputs_folder)
    else:
        outputs_path = Path.cwd() / "outputs"
        if not outputs_path.exists():
            outputs_path = Path.cwd()

    # Check if the outputs folder contains a scenarios.json file, i.e. that it is a valid outputs folder
    scenario_files = outputs_path.glob("**/scenarios.json")
    if not any(scenario_files):
        raise FileNotFoundError(
            f"No scenarios.json files found in the outputs folder: {outputs_path}. "
            "Please provide a valid outputs folder using '-o <path-to-folder>'."
        )
    return str(outputs_path)


def parse_arguments_and_run() -> None:
    parser = ArgumentParser(
        description="ZEN Temple - Visualization web platform for ZEN Garden"
    )

    group = parser.add_argument_group("Server Options")
    group.add_argument(
        "-p",
        "--port",
        required=False,
        type=int,
        default=8000,
        help="port on which to run the local server",
    )
    group.add_argument(
        "-o",
        "--outputs-folder",
        required=False,
        type=str,
        default=None,
        help="path to your solutions folder. Per default looks for data in ./outputs or in the current working directory",
    )
    group.add_argument(
        "--significant-digits",
        required=False,
        type=int,
        default=None,
        help="number of significant digits to use in the response. If not set, uses the value from the environment variable RESPONSE_SIGNIFICANT_DIGITS (default: 4)",
    )

    group = parser.add_argument_group("Developer Options")
    group.add_argument(
        "--app-name",
        required=False,
        type=str,
        default="",
        help="name of the app",
    )
    group.add_argument(
        "--api-url",
        required=False,
        type=str,
        default=None,
        help="URL to the API to fetch the data from",
    )
    group.add_argument(
        "--reload",
        required=False,
        action="store_true",
        help="enable reload for development purposes",
    )
    group.add_argument(
        "--no-open-browser",
        required=False,
        action="store_true",
        help="do not open the browser automatically",
    )
    group.add_argument(
        "--fd",
        required=False,
        type=int,
        default=None,
        help="file descriptor for the server that the server can bind to",
    )
    args = parser.parse_args()

    outputs_folder = find_outputs_folder(args.outputs_folder)

    start_server(
        outputs_folder,
        args.port,
        app_name=args.app_name,
        api_url=args.api_url,
        significant_digits=args.significant_digits,
        reload=args.reload,
        no_open_browser=args.no_open_browser,
        fd=args.fd,
    )


if __name__ == "__main__":
    parse_arguments_and_run()



from typing import Any, Optional

from fastapi import Query, APIRouter

from ..models.solution import SolutionDetail, SolutionList
from ..repositories.solution_repository import solution_repository

router = APIRouter(prefix="/solutions", tags=["Solutions"])


@router.get("/list")
async def get_list() -> list[SolutionList]:
    """
    Get a list of the available solutions.
    """
    return solution_repository.get_list()


@router.get("/get_detail/{solution_name}")
async def get_detail(solution_name: str) -> SolutionDetail:
    """
    Get the details of a solution.
    """
    return solution_repository.get_detail(solution_name)


@router.get("/get_total")
async def get_total(
    solution_name: str,
    components: list[str] = Query(...),
    unit_component: Optional[str] = None,
    scenario: Optional[str] = None,
) -> dict[str, Optional[str]]:
    """
    Get the total of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    return solution_repository.get_total(
        solution_name, ",".join(components), unit_component, scenario
    )


@router.get("/get_full_ts")
async def get_full_ts(
    solution_name: str,
    components: list[str] = Query(...),
    unit_component: Optional[str] = None,
    scenario: Optional[str] = None,
    year: Optional[int] = None,
    rolling_average_size: int = 1,
) -> dict[str, Optional[list[dict[str, Any]] | str]]:
    """
    Get the total of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    return solution_repository.get_full_ts(
        solution_name,
        ",".join(components),
        unit_component,
        scenario,
        year,
        rolling_average_size,
    )


@router.get("/get_unit/{solution_name}/{variable_name}")
async def get_unit(solution_name: str, variable_name: str) -> Optional[str]:
    """
    Get the unit of a variable given the solution name, the variable name, and the scenario. If no scenario is provided, the first scenarios in the list is taken.
    """
    return solution_repository.get_unit(solution_name, variable_name)


@router.get("/get_energy_balance/{solution_name}/{node_name}/{carrier_name}")
async def get_energy_balance(
    solution_name: str,
    node_name: str,
    carrier_name: str,
    scenario: Optional[str] = None,
    year: Optional[int] = 0,
    rolling_average_size: int = 1,
) -> dict[str, list[dict[str, Any]]]:
    """
    Get the energy balance of a specific node and carrier given the solution name, the node name, the carrier, the scenario, and the year.
    If no scenario and/or year is provided, the first one is taken.
    """
    return solution_repository.get_energy_balance(
        solution_name, node_name, carrier_name, scenario, year, rolling_average_size
    )



import os
from functools import cache
from os import walk
from typing import Any, Optional

import pandas as pd
import numpy as np
from fastapi import HTTPException
from zen_garden.postprocess.results import Results  # type: ignore

from zen_temple.utils import get_variable_name

from ..config import config
from ..models.solution import (
    SolutionDetail,
    SolutionList,
)


class SolutionRepository:
    def __load_results(self, solution_name: str) -> Results:
        """
        Loads the results of a solution given its name.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :return: Results object of the solution.
        """
        path = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        return Results(path)

    def __dataframe_to_csv(self, df: "pd.DataFrame | pd.Series[Any]") -> str:
        """
        Converts a DataFrame or Series to a CSV string.
        """
        return df.to_csv(
            lineterminator="\n",
            float_format=f"%.{config.RESPONSE_SIGNIFICANT_DIGITS}g",
        )

    def get_list(self) -> list[SolutionList]:
        """
        Creates a list of Solution-objects of all solutions that are contained in any folder contained in the configured SOLUTION_FOLDER.

        This function is very forgiving, it tries to instantiate a solution for all folders in SOLUTION_FOLDER that contain a 'scenarios.json' file.
        If this fails, it skips the folder.
        """
        solutions_folders: set[str] = set()
        ans = []
        # TODO this is bad because if you accidentally have a scenarios.json in a subscenario folder, it will be included in the list.
        #      Better check if the parent folder is a solution, i.e., whether is has a scenarios.json
        for dirpath, dirnames, filenames in walk(config.SOLUTION_FOLDER):
            if "scenarios.json" in filenames:
                solutions_folders.add(dirpath)
                # Prevent os.walk from going deeper into this folder
                dirnames.clear()
        for folder in solutions_folders:
            try:
                ans.append(SolutionList.from_path(folder))
            except (FileNotFoundError, NotADirectoryError) as e:
                print(str(e) + f" - Skip {folder}")
                continue
        return ans

    @cache
    def get_detail(self, solution_name: str) -> SolutionDetail:
        """
        Returns the SolutionDetail of a solution given its name.

        The solution name can contain dots which are treated as folders.
        So for example foo/bar.solution will resolve to the solution contained in foo/bar/solution, relative to
        the SOLUTION_FOLDER config value.

        :param solution_name: Name of the solution
        """
        path = os.path.join(config.SOLUTION_FOLDER, *solution_name.split("."))
        return SolutionDetail.from_path(path)

    @cache
    def get_full_ts(
        self,
        solution_name: str,
        components_str: str,
        unit_component: Optional[str] = None,
        scenario: Optional[str] = None,
        year: Optional[int] = None,
        rolling_average_window_size: int = 1,
    ) -> dict[str, Optional[list[dict[str, Any]] | str]]:
        """
        Returns the full ts and the unit of a component given the solution name, the component name and the scenario name.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param component: Name of the component.
        :param scenario: Name of the scenario. If skipped, the first scenario is taken.
        :param year: The year of the ts. If skipped, the first year is taken.
        """
        components = [x for x in components_str.split(",") if x != ""]
        if len(components) == 0:
            raise HTTPException(status_code=400, detail="No components provided!")

        results = self.__load_results(solution_name)
        unit = self.__read_out_units(
            results,
            (
                unit_component
                if unit_component and unit_component is not None
                else components[0]
            ),
        )
        response: dict[str, Optional[list[dict[str, Any]] | str]] = {"unit": unit}

        if year is None:
            year = results.get_analysis(scenario).earliest_year_of_data

        for component in components:
            full_ts = results.get_full_ts(component, scenario_name=scenario, year=year)
            if full_ts.shape[0] == 0:
                response.update({component: []})
                continue

            full_ts = full_ts[~full_ts.index.duplicated(keep="first")]
            full_ts = full_ts.loc[
                (abs(full_ts) > config.EPS * max(full_ts)).any(axis=1)
            ]

            if rolling_average_window_size > 1:
                full_ts = self.__compute_rolling_average(
                    full_ts, rolling_average_window_size
                )

            res = self.__quantify_response(full_ts)
            response[component] = res

        return response

    def __compute_rolling_average(
        self, df: "pd.DataFrame | pd.Series[Any]", window_size: int
    ) -> "pd.DataFrame | pd.Series[Any]":
        if df.shape[0] == 0:
            return df

        # Append end of df to beginning
        df = df[df.columns[-window_size:].to_list() + df.columns.to_list()]

        # Compute rolling average
        df = df.T.rolling(window_size).mean().dropna().T

        # Rename columns so it starts at 0
        df = df.set_axis(range(df.shape[1]), axis=1)

        return df

    def __quantify_response(self, df: "Any") -> list[dict[str, Any]]:
        """
        Converts a DataFrame or Series to a dictionary with quantized values.
        Quantization is done by mapping the values of each row to the interval [0, quantile),
        converting them to integers and delta encode them.

        The response contains the transformation parameters `(translation, scale)`
        such that we can reverse this process using:

        ```
        values = np.cumsum(values)
        values = values * scale + translation
        ```

        This design is analogous to TopoJSON's quantization scheme.
        """
        if df.shape[0] == 0:
            return []

        # Get index and data values
        index_names = df.index.names
        index_values = df.index.to_numpy()
        data_values = df.to_numpy()

        # Compute min/max per row
        min_values = data_values.min(axis=1)
        max_values = data_values.max(axis=1)
        diff_values = max_values - min_values

        # Compute translation and scale parameters for mapping the value to [0, quantile)
        translations = min_values
        quantile = 10**(config.RESPONSE_SIGNIFICANT_DIGITS)
        scales = (diff_values + config.EPS) / (quantile - 1)

        # Apply translation and scaling
        data_values = (data_values - translations[:, None]) / scales[:, None]

        # Convert to int
        data_values = data_values.astype(int)

        # Delta encode values
        data_values = np.diff(data_values, prepend=0)

        return [
            {
                **dict(zip(index_names, idx)),
                "d": row.tolist(),
                "t": (translation, scale)
            }
            for idx, row, translation, scale in zip(index_values, data_values, translations, scales)
        ]

    @cache
    def get_total(
        self,
        solution_name: str,
        components_str: str,
        unit_component: Optional[str] = None,
        scenario: Optional[str] = None,
    ) -> dict[str, Optional[str]]:
        """
        Returns the total and the unit of a component given the solution name, the scenario name and the component name.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param component: Name of the component.
        :param scenario: Name of the scenario. If skipped, the first scenario is taken.
        """
        components = [x for x in components_str.split(",") if x != ""]
        if len(components) == 0:
            raise HTTPException(status_code=400, detail="No components provided!")

        results = self.__load_results(solution_name)
        unit = self.__read_out_units(
            results,
            (
                unit_component
                if unit_component and unit_component is not None
                else components[0]
            ),
        )
        response = {"unit": unit}

        for component in components:
            try:
                total: pd.DataFrame | pd.Series[Any] = results.get_total(
                    component, scenario_name=scenario
                )
            except KeyError:
                raise HTTPException(
                    status_code=404, detail=f"{component} not found in {solution_name}"
                )

            # Skip irrelevant rows in dataframes
            if type(total) is not pd.Series:
                total = total.loc[(abs(total) > config.EPS * max(total)).any(axis=1)]
            response.update({component: self.__dataframe_to_csv(total)})

        return response

    def get_unit(self, solution_name: str, component: str) -> Optional[str]:
        """
        Returns the unit of a component given the solution name. If there are several units in the requested component, it returns it in form of a CSV string.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        """
        return self.__read_out_units(self.__load_results(solution_name), component)

    def __read_out_units(self, results: Results, component: str) -> Optional[str]:
        """
        Reads out the units of a component from the results object.
        """
        try:
            unit = results.get_unit(component, convert_to_yearly_unit=True)
            if type(unit) is str:
                unit = pd.DataFrame({0: [unit]})
            return self.__dataframe_to_csv(unit)
        except Exception as e:
            print(e)
            return None

    @cache
    def get_energy_balance(
        self,
        solution_name: str,
        node: str,
        carrier: str,
        scenario: Optional[str] = None,
        year: Optional[int] = None,
        rolling_average_window_size: int = 1,
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Returns the energy balance dataframes of a solution.
        It drops duplicates of all dataframes and removes the variables that only contain zeros.

        :param solution_name: Name of the solution. Dots will be regarded as subfolders (foo.bar => foo/bar).
        :param node: The name of the node.
        :param carrier: The name of the carrier.
        :param scenario: The name of the scenario. If skipped, the first scenario is taken.
        :param year: The desired year. If skipped, the first year is taken.
        :param rolling_average_window_size: Size of the rolling average window.
        """
        results = self.__load_results(solution_name)

        if year is None:
            year = 0

        balances: dict[str, pd.DataFrame | pd.Series[Any]] = (
            results.get_energy_balance_dataframes(node, carrier, year, scenario)
        )

        # Drop duplicates of all dataframes
        balances = {
            key: val[~val.index.duplicated(keep="first")]
            for key, val in balances.items()
        }

        # Drop variables that only contain zeros (except for demand)
        for key, series in balances.items():
            demand_name = get_variable_name(
                "demand", results.get_analysis().zen_garden_version
            )

            if type(series) is not pd.Series and key != demand_name:
                if series.empty:
                    continue
                balances[key] = series.loc[
                    (abs(series) > config.EPS * max(series)).any(axis=1)
                ]

            if rolling_average_window_size > 1:
                balances[key] = self.__compute_rolling_average(
                    balances[key], rolling_average_window_size
                )

        ans = {key: self.__quantify_response(val) for key, val in balances.items()}

        return ans


solution_repository = SolutionRepository()



import os

from dotenv import load_dotenv

load_dotenv()


class Config:
    def __init__(self) -> None:
        self.SOLUTION_FOLDER: str = os.getenv("SOLUTION_FOLDER", "./outputs")
        self.EPS = float(os.getenv("EPS", 1e-6))
        self.RESPONSE_SIGNIFICANT_DIGITS = int(os.getenv("RESPONSE_SIGNIFICANT_DIGITS", 4))

        self.check()

    def check(self) -> None:
        for key, val in self.__dict__.items():
            if val is None:
                raise Exception(f"Env-Variable {key} is missing!")


config = Config()



from typing import Optional


class Version:
    def __init__(self, version_string: str):
        sub_versions: list[int] = [int(i) for i in version_string.split(".")]
        self.versions = sub_versions
        self.major = sub_versions[0]
        self.minor = sub_versions[1]
        self.patch = sub_versions[2]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return NotImplemented

        for i in range(len(self.versions)):
            if self.versions[i] != other.versions[i]:
                return False

        return True

    def __gt__(self, other: "Version") -> bool:
        for i in range(len(self.versions)):
            if self.versions[i] != other.versions[i]:
                return self.versions[i] > other.versions[i]
        return False

    def __repr__(self) -> str:
        return ".".join([str(i) for i in self.versions])


variable_versions = {
    "0.0.0": {},
    "1.9.0": {
        "capex_yearly": "cost_capex_yearly",
        "opex_yearly": "cost_opex_yearly",
        "cost_opex_total": "cost_opex_yearly_total",
        "cost_opex": "cost_opex_variable",
        "cost_capex_total": "cost_capex_yearly_total",
        "cost_capex": "cost_capex_overnight",
    },
}


def get_variable_name(variable: str, version_string: Optional[str] = None) -> str:
    if version_string is None:
        version_string = "0.0.0"

    results_version = Version(version_string)
    relevant_version_string = "0.0.0"

    for defined_version_string in variable_versions:
        defined_version = Version(defined_version_string)

        if defined_version > results_version:
            break
        relevant_version_string = defined_version_string

    if variable in variable_versions[relevant_version_string]:
        return variable_versions[relevant_version_string][variable]

    return variable
